;; Generated package description from inspector.el  -*- no-byte-compile: t -*-
(define-package "inspector" "0.39" "Tool for inspection of Emacs Lisp objects" '((emacs "27.1")) :commit "52a64993ac36ed3ed0be51b6a0d54d190edc9c74" :authors '(("Mariano Montone" . "marianomontone@gmail.com")) :maintainer '("Mariano Montone" . "marianomontone@gmail.com") :keywords '("debugging" "tool" "lisp" "development") :url "https://github.com/mmontone/emacs-inspector")
